from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarFilesOverride(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.filesoverride',
                         severity='warning',
                         message="'{a}' should not be overridden")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue='FILES')
        for i in items:
            if not i.AppendOperation():
                res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(a=i.VarName))
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable, Function
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import Constants

from oelint_adv.cls_rule import Rule, Classification


class VarSpecNonCore(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.noncoreoverride',
                         severity='error',
                         message="'{a}' need to be specific to a custom MACHINE or DISTRO to be Yocto project compliant",
                         run_on=[Classification.BBAPPEND])

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        self._core_constants = Constants(self._state.core_mod_ref)

        res = []
        items: List[Variable] = stash.GetItemsFor(
            filename=_file, classifier=(Variable.CLASSIFIER, Function.CLASSIFIER))
        for i in items:
            subs = [x for x in i.SubItems if x not in [
                'append', 'prepend', 'remove', '${PN}', '${BPN}']]
            if isinstance(i, Variable):
                name = i.VarName
                remaining_subs = [
                    x for x in subs if x not in self._core_constants.VariablesKnown]
            elif isinstance(i, Function):  # pragma: no cover
                name = i.FuncName
                remaining_subs = [
                    x for x in subs if x not in self._core_constants.FunctionsKnown]
            remaining_subs = [
                x for x in remaining_subs if x not in self._core_constants.DistrosKnown]
            remaining_subs = [
                x for x in remaining_subs if x not in self._core_constants.MachinesKnown]

            if not remaining_subs:
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg=self.Msg.format(a=name))

        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarsDownloadfilename(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.downloadfilename',
                         severity='warning',
                         message="Fetcher does create a download artifact without 'PV' in the filename")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')

        _pv = stash.ExpandTerm(_file, '${PV}')

        for i in items:
            for item in [stash.ExpandTerm(_file, x) for x in i.get_items()]:
                _scrcomp = stash.GetScrComponents(item)
                if _scrcomp.get('scheme', '') in ['https', 'http', 'ftp']:
                    _src = _scrcomp.get('src', '')
                    _dfn = _scrcomp.get('options', {}).get(
                        'downloadfilename', '')

                    if _pv not in _src:
                        if _dfn and _pv in _dfn:
                            continue
                        res += self.finding(i.Origin, i.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarBugtrackerIsUrl(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.fileextrapathsop',
                         severity='error',
                         message="'FILESEXTRAPATHS' should only be used in combination with ' := '")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='FILESEXTRAPATHS')
        for i in items:
            if i.VarOp.strip() != ':=':
                res += self.finding(i.Origin, i.InFileLine)
        return res



import os
from typing import List, Tuple

from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule, Classification


class FileNoSpaces(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.file.underscores',
                         severity='error',
                         message='FOO',
                         run_on=[Classification.RECIPE])

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _basename, _ext = os.path.splitext(os.path.basename(_file))
        if _ext in ['.bb']:  # pragma: no cover
            if stash.IsPackageGroup(_file) or stash.IsImage(_file):
                return []
            _us = [x for x in _basename if x == '_']
            if len(_us) > 1:
                res += self.finding(_file, 1,
                                    override_msg="Filename should not contain more than one '_'")
            elif not _us:
                res += self.finding(
                    _file, 1, override_msg="Filename should contain at least one '_' in the end")
        return res



import os
from difflib import SequenceMatcher
import functools
from typing import List, Tuple

from oelint_parser.cls_item import FlagAssignment, Function, Item, Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule


class VarMisspell(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.mispell',
                         severity='warning',
                         message='<FOO>',
                         appendix=['unknown'])
        try:
            self._minconfidence = float(os.environ.get(
                'OELINT_MISSPELL_CONFIDENCE', 'not-a-float'))
        except ValueError:
            self._minconfidence = 0.8

        self._layername_extensions_on = [
            'BBFILE_PATTERN',
            'BBFILE_PRIORITY',
            'BBFILE_PATTERN_IGNORE_EMPTY',
            'LAYERDEPENDS',
            'LAYERRECOMMENDS',
            'LAYERSERIES_COMPAT',
            'LAYERVERSION',
        ]

    @functools.cache  # noqa: B019
    def get_best_match(self, item: str, _list: List[str]) -> str:
        _dict = sorted([(SequenceMatcher(None, item, k).ratio(), k)
                        for k in _list], key=lambda x: x[0], reverse=True)
        if _dict and _dict[0][0] >= self._minconfidence:
            return _dict[0][1]
        return ''

    def get_collection_vars(self, _file, stash):
        _bbfile_collections = stash.ExpandVar(_file, attribute=Variable.ATTR_VAR,
                                              attributeValue='BBFILE_COLLECTIONS').get('BBFILE_COLLECTIONS', [])
        if not hasattr(self, '_collection_vars') or self._collection_vars != _bbfile_collections:

            self._collection_vars = _bbfile_collections
            res = set()
            for collection in _bbfile_collections:
                for var in self._layername_extensions_on:  # noqa: VNE002
                    res.add(f'{var}_{collection}')

        return res  # pragma: no cover

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []

        _all: List[Item] = stash.GetItemsFor(filename=_file)
        items: List[Item] = stash.Reduce(_all, filename=_file, classifier=[
                                         Variable.CLASSIFIER, FlagAssignment.CLASSIFIER])
        _extras = [f'SRCREV_{x}' for x in stash.GetValidNamedResources(_file)]
        _pkgs = [x for x in stash.GetValidPackageNames(
            _file, strippn=True) if x]
        _taskname = CONSTANTS.FunctionsKnown + \
            [x.FuncName for x in _all if isinstance(x, Function)]
        _vars = CONSTANTS.VariablesKnown

        for i in items:
            if isinstance(i, Variable):
                _cleanvarname = i.VarName
                if _cleanvarname.startswith('_'):
                    continue
                if _cleanvarname in _vars:
                    continue
                if _cleanvarname in _extras:
                    continue
                for pkg in _pkgs:
                    if _cleanvarname.endswith(pkg):
                        _cleanvarname = ''.join(
                            _cleanvarname.rsplit(pkg, 1))  # pragma: no cover
            else:
                _cleanvarname = i.VarName
                if _cleanvarname.startswith('_'):
                    continue
                if _cleanvarname in _taskname:
                    continue
                if _cleanvarname in _vars:
                    continue
                if _cleanvarname in _extras:
                    continue
            # Does it even make sense to involve collection related variables?
            if any(i.VarName.startswith(v) for v in self._layername_extensions_on):
                # We tried to delay calling ExpandVar for as long as possible, but it is time.
                _vars.extend(self.get_collection_vars(_file, stash))
                if i.VarName in _vars:
                    continue
            _used = False
            for a in _all:
                if a == i:
                    continue
                if '${{{a}}}'.format(a=i.VarName) in a.Raw \
                        or 'getVar("{a}"'.format(a=i.VarName) in a.Raw \
                        or "getVar('{a}'".format(a=i.VarName) in a.Raw:
                    _used = True
                    break
            if _used:
                continue
            _bestmatch = self.get_best_match(_cleanvarname, frozenset(_vars))
            if _bestmatch:
                res += self.finding(i.Origin, i.InFileLine,
                                    "'{a}' is unknown, maybe you meant '{b}'".format(
                                        a=_cleanvarname, b=_bestmatch))
            else:
                res += self.finding(i.Origin, i.InFileLine,
                                    "'{a}' is unknown. Consider adding it via --constantmods".format(
                                        a=_cleanvarname),
                                    appendix='unknown',
                                    severity_override='info')
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarUnneededFilesSetting(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.filessetting',
                         severity='warning',
                         message='Check for improvable FILES settings',
                         appendix=['hidden', 'double'])

    def __find_match_from_stash(self,
                                _files: List[Variable],
                                variable_: str,
                                needle: str,
                                msg: str,
                                appendix: str,
                                onappendonly: bool = False) -> List[Tuple[str, int, str]]:
        res = []
        for i in _files:
            if variable_ in i.SubItems and 'remove' not in i.SubItems and needle in i.VarValue:  # pragma: no cover
                if (onappendonly and i.IsAppend()) or (not onappendonly):
                    res += self.finding(i.Origin, i.InFileLine,
                                        override_msg=msg, appendix=appendix)
        return res

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _all_files = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                       attribute=Variable.ATTR_VAR, attributeValue='FILES')
        if not any(_all_files):
            return res

        _expanded = stash.ExpandVar(filename=_file, attribute=Variable.ATTR_VAR)
        override_delimiter = _all_files[-1].OverrideDelimiter
        _seenpath = {}
        for p in _expanded['PACKAGES']:
            _files = 'FILES{o}{a}'.format(o=override_delimiter, a=p)
            _convfiles = _files.replace(_expanded['PN'][0], '${PN}')
            if _files in _expanded:
                _pattern = _expanded[_files]
                for _p in _pattern:
                    # double setting in FILES
                    # try to find with both unexpanded and expanded values
                    if len([x for x in _pattern if x == _p]) > 1:
                        res += self.__find_match_from_stash(_all_files, p.replace(_expanded['PN'][0], '${PN}'), _p,
                                                            '{a} is already set by default or in this recipe'.format(a=_p), 'double', True)
                        res += self.__find_match_from_stash(_all_files, p, _p,
                                                            '{a} is already set by default or in this recipe'.format(a=_p), 'double', True)
                    # useless as hidden by previous package
                    if _p in _seenpath.keys() and _seenpath[_p] != _convfiles:
                        # try to find with both unexpanded and expanded values
                        res += self.__find_match_from_stash(_all_files, p.replace(_expanded['PN'][0], '${PN}'), _p,
                                                            '{a} is already covered by {b}'.format(a=_p, b=_seenpath[_p]),
                                                            'hidden')
                        res += self.__find_match_from_stash(_all_files, p, _p,
                                                            '{a} is already covered by {b}'.format(a=_p, b=_seenpath[_p]),
                                                            'hidden')
                    _seenpath[_p] = _convfiles
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarDescSame(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.descriptionsame',
                         severity='warning',
                         message="'DESCRIPTION' is the same a 'SUMMARY' - it can be removed then")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='DESCRIPTION')
        if not any(items):
            return res

        summary = ''.join(stash.ExpandVar(_file, attribute=Variable.ATTR_VAR, attributeValue='SUMMARY').get('SUMMARY', ['']))
        desc = ''.join(stash.ExpandVar(_file, attribute=Variable.ATTR_VAR,
                       attributeValue='DESCRIPTION').get('DESCRIPTION', ['']))
        if desc.strip() == summary.strip():
            res += self.finding(items[0].Origin, items[0].InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable, FlagAssignment
from oelint_parser.cls_stash import Stash
from oelint_parser.parser import INLINE_BLOCK

from oelint_adv.cls_rule import Rule


class VarSRCURIMutualExItems(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.srcurimutualex',
                         severity='error',
                         message="'SRCREV' and a checksum are defined for the same item - only one is applicable")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _srcuri: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                    attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        _map = {}
        for item in _srcuri:
            for value in [x.strip("'") for x in item.get_items() if x and INLINE_BLOCK not in x]:
                components = stash.GetScrComponents(value)

                _map[components['options'].get('name', '')] = {
                    'type': components['scheme']}

        _srcrevs: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                     attribute=Variable.ATTR_VAR,
                                                     attributeValue=['SRCREV'] + [f'SRCREV_{k}' for k in _map.keys() if k])
        _flags: List[Variable] = stash.GetItemsFor(filename=_file, classifier=FlagAssignment.CLASSIFIER,
                                                   attribute=FlagAssignment.ATTR_NAME,
                                                   attributeValue='SRC_URI')

        for item in _srcrevs:
            needle = item.VarNameCompleteNoModifiers.split('_')[-1]
            if needle == 'SRCREV':
                needle = ''
            if needle in _map:
                _map[needle]['srcrev'] = item

        for item in _flags:
            if item.Flag.endswith('sum'):
                needle = item.Flag.split('.')[0]
                _sum = item.Flag.split('.')[-1]
                if needle.endswith('sum'):
                    needle = ''
                if needle in _map:
                    _map[needle][_sum] = item

        for _, v in _map.items():
            if 'srcrev' not in v:
                continue
            for sk, sv in v.items():
                if not sk.endswith('sum'):
                    continue
                res += self.finding(sv.Origin, sv.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Function
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarDUsageInPkgfunc(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.dusageinpkgfuncs',
                         severity='error',
                         message='Use $D instead of ${{D}} in {func}')

    def __getMatches(self, _file: str, stash: Stash) -> List[Function]:
        res = []
        items: List[Function] = stash.GetItemsFor(filename=_file,
                                                  classifier=Function.CLASSIFIER,
                                                  attribute=Function.ATTR_FUNCNAME,
                                                  attributeValue=[
                                                      'pkg_preinst',
                                                      'pkg_postinst',
                                                      'pkg_postrm',
                                                      'pkg_prerm',
                                                  ],
                                                  )
        for i in items:
            if '${D}' in i.FuncBodyRaw:
                res.append(i)
        return res

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(func=i.FuncName))
        return res

    def fix(self, _file: str, stash: Stash) -> List[str]:
        res = []
        for i in self.__getMatches(_file, stash):
            i.RealRaw = i.RealRaw.replace('${D}', '$D')
            i.Raw = i.Raw.replace('${D}', '$D')
            res.append(_file)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Function, TaskAdd
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule


class TaskAddNoTaskBody(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.task.addnotaskbody',
                         severity='warning',
                         message="The added task '{FUNC}' is not existing or has no body")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=TaskAdd.CLASSIFIER):
            if item.FuncName in CONSTANTS.FunctionsOrder:
                # not for builtin types
                continue
            if item.FuncName in [x.replace('do_', '', 1) for x in CONSTANTS.FunctionsOrder]:
                # not for builtin types - probed for missing prefix
                continue
            _ta = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                    attribute='FuncName')
            _filt = [x for x in _ta or [] if not isinstance(
                x, str) and x.FuncName == item.FuncName]
            _filt += [x for x in _ta or []
                      if not isinstance(x, str) and x.FuncName == 'do_' + item.FuncName]
            if not any(_filt):
                res += self.finding(item.Origin, item.InFileLine,
                                    self.Msg.replace('{FUNC}', item.FuncName))
            elif not any(x for x in _filt if x.FuncBodyStripped):
                res += self.finding(item.Origin, item.InFileLine,
                                    self.Msg.replace('{FUNC}', item.FuncName))
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Function, Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS
from oelint_parser.rpl_regex import RegexRpl

from oelint_adv.cls_rule import Rule


class VarPnBpnUsage(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.func.specific',
                         severity='error',
                         message="'{func}' is set specific to ['{b}'], but isn't known from PACKAGES, MACHINE, DISTRO")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                  attribute=Function.ATTR_FUNCNAME)
        _comp = ''.join(x.VarValueStripped for x in stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                                      attribute=Variable.ATTR_VAR,
                                                                      attributeValue='COMPATIBLE_MACHINE'))
        _packages = stash.GetValidPackageNames(_file)
        _machines = CONSTANTS.MachinesKnown
        _distros = CONSTANTS.DistrosKnown
        _builtin_funcs = ['pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm', 'ptest']
        _operations = ['append', 'prepend', 'remove']
        for i in items:
            subs = [(x, stash.ExpandTerm(_file, x)) for x in i.SubItems]
            for subitem in subs:
                sub, expanded = subitem
                if (expanded in _operations) or (sub in _operations):
                    continue  # pragma: nocover_3.9 - coverage looks buggy on 3.9
                if (expanded in _distros) or (sub in _distros):
                    continue  # pragma: nocover_3.9 - coverage looks buggy on 3.9
                if (expanded in _packages) or (sub in _packages):
                    continue  # pragma: nocover_3.9 - coverage looks buggy on 3.9
                if (expanded in _machines) or (sub in _machines):
                    continue  # pragma: nocover_3.9 - coverage looks buggy on 3.9
                if (expanded in _builtin_funcs) or (sub in _builtin_funcs):
                    continue  # pragma: nocover_3.9 - coverage looks buggy on 3.9
                if _comp and (RegexRpl.match(_comp, expanded) or RegexRpl.match(_comp, sub)):
                    continue  # pragma: nocover_3.9 - coverage looks buggy on 3.9
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg=self.Msg.format(func=i.FuncName, b=sub))
        return res



import ast
import textwrap
from typing import List, Tuple

from oelint_parser.cls_item import Function
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class TaskPythonPrefix(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.task.pythonprefix',
                         severity='warning',
                         message='Tasks containing python code, should be prefixed with python in function header')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Function] = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER)
        for item in items:
            # Don't make assumptions about functions that consist of single
            # line only
            if len([x for x in item.FuncBodyRaw.split('\n') if x.strip()]) < 2:
                continue
            try:
                ast.parse(textwrap.dedent(item.FuncBodyRaw.rstrip('}\n')), 'tempfile')
                if not item.IsPython:
                    res += self.finding(item.Origin, item.InFileLine)
            except Exception:  # noqa: S110
                pass  # noqa: S110 - intentionally ignore all errors
        return res



from typing import List, Tuple

from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class FileNoSpaces(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.file.nospaces',
                         severity='error',
                         message='Path to file contains spaces. Please remove them')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _layer_root = stash.GetLayerRoot(_file)
        _relpath = _file.replace(_layer_root, '').lstrip('/')
        if ' ' in _relpath:
            res += self.finding(_file, 1)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarMultiLineIndent(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.multilineident',
                         severity='info',
                         message='On a multiline assignment, line indent is desirable. {a} set, {b} desirable')

    def __line_stats(self, i: Variable) -> Tuple[int, int, List[Tuple[int, int, str]]]:
        _map = []

        _lines = i.VarValueStripped.replace('\x1b"', '"').split('\x1b')
        non_empty_line_indent = 4
        first_line_has_content = False
        last_line_indent = 0
        if _lines[0].strip():
            first_line_has_content = True
            non_empty_line_indent = i.Raw.index(_lines[0])
            last_line_indent = non_empty_line_indent

        for index, value in enumerate(_lines):
            if value.strip(' \x1b'):
                if first_line_has_content and index == 0:
                    _map.append((0, len(value) - len(value.lstrip()), value.lstrip()))
                else:
                    _map.append((non_empty_line_indent, len(value) - len(value.lstrip()),
                                " " * non_empty_line_indent + value.lstrip()))
            else:
                _map.append((0, 0, value.lstrip()))

        return (last_line_indent, non_empty_line_indent, _map)

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if not i.IsMultiLine():
                continue
            _, _, _indent_map = self.__line_stats(i)
            for index, _line in enumerate(_indent_map):
                expected, actual, _ = _line
                if expected != actual:
                    res += self.finding(i.Origin, i.InFileLine + index,
                                        self.format_message(a=actual, b=expected), blockoffset=i.InFileLine)
        return res

    def fix(self, _file: str, stash: Stash) -> List[str]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if not i.IsMultiLine():
                continue
            _lines = i.VarValueStripped.split('\x1b')
            _last_line_indent, _standard_indent, _indent_map = self.__line_stats(i)

            def _rreplace(in_: str, needle: str, repl: str) -> str:
                return in_[::-1].replace(needle[::-1], repl[::-1], 1)[::-1]

            fix_applied = False
            for index, value in enumerate(_lines):
                if value != _indent_map[index][2]:
                    if not value.strip():
                        continue
                    replacement = _indent_map[index][2]
                    fix_applied = True
                    # Note: we need to start replacing from the back of the string
                    # as otherwise a last line containing only whitespace
                    # will affect prior lines
                    i.VarValue = _rreplace(i.VarValue, value, replacement)
                    i.Raw = _rreplace(i.Raw, value, replacement)
                    i.RealRaw = _rreplace(i.RealRaw, value, replacement)

            _last_line = i.VarValue.split('\x1b')[-1]
            _last_line_stripped = _last_line.strip()
            if len(_last_line_stripped) > 1:
                _last_line_expected = f'{" " * _standard_indent + _last_line_stripped}'
            else:
                _last_line_expected = f'{" " * _last_line_indent + _last_line_stripped}'

            if _last_line != _last_line_expected:
                if _last_line.strip('"\''):
                    i.VarValue = _rreplace(i.VarValue, _last_line, _last_line_expected)
                    i.Raw = _rreplace(i.Raw, _last_line, _last_line_expected)
                    i.RealRaw = _rreplace(i.RealRaw, _last_line, _last_line_expected)
                    fix_applied = True

            if fix_applied:
                res.append(_file)

        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule, Classification


class VarQuoted(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.append.protvars',
                         severity='error',
                         message="Variable '{VAR}' shouldn't be set as part of a bbappend",
                         run_on=[Classification.BBAPPEND],
                         appendix=CONSTANTS.GetByPath('variables/protected-append'))

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in [x for x in items if x.VarName in CONSTANTS.GetByPath('variables/protected-append')]:
            if i.VarOp.strip() not in ['??=', '?=']:
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg=self.Msg.replace('{VAR}', i.VarName), appendix=i.VarName)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Include
from oelint_parser.cls_stash import Stash
from oelint_parser.rpl_regex import RegexRpl

from oelint_adv.cls_rule import Rule


class VarMultiInclude(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.multiinclude',
                         severity='warning',
                         message="'{INC}' is included multiple times")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Include] = stash.GetItemsFor(filename=_file, classifier=Include.CLASSIFIER)
        _map = {}
        for i in items:
            for incname in RegexRpl.split(r'\s|,', i.IncName):
                if not incname or not incname.strip():
                    continue  # pragma: no cover
                key = i.FileIncluded if i.FileIncluded else incname.strip()
                if key not in _map:
                    _map[key] = {'key': incname.strip(), 'value': 0, 'origin': i}
                _map[key]['value'] += 1
        for _, value in _map.items():
            if value.get('value', 0) > 1:
                res += self.finding(value.get('origin').Origin,
                                    value.get('origin').InFileLine,
                                    self.Msg.replace('{INC}', value.get('key')))
        return res



from typing import List, Tuple

from oelint_parser.cls_item import FlagAssignment, Function
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule


class TaskNetwork(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.task.network',
                         severity='warning',
                         message="Task '{FUNC}' uses the network flag, that can lead to unpredictable results",
                         valid_from_release='kirkstone')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        task_names = {x.FuncName for x in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER)}
        task_names.update(CONSTANTS.FunctionsKnown)
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=FlagAssignment.CLASSIFIER,
                                      attribute=FlagAssignment.ATTR_NAME,
                                      attributeValue=task_names):
            if item.VarName in ['do_fetch']:
                continue
            if item.Flag in ['network'] and item.ValueStripped not in ['0']:
                res += self.finding(item.Origin, item.InFileLine,
                                    self.Msg.replace('{FUNC}', item.VarName))
        return res
